from typing import List, Tuple

from allennlp.common import Registrable
from overrides import overrides

class FEVERInstanceGenerator(Registrable):

    def generate_instances(self, reader, evidence:List[List[Tuple[int, str, int]]], claim:str):
        raise NotImplemented("This preprocessing function should be implemented")


@FEVERInstanceGenerator.register("concatenate")
class ConcatenateEvidence(FEVERInstanceGenerator):

    @staticmethod
    def _flatten(l):
        return [item for sublist in l for item in sublist]

    @overrides
    def generate_instances(self, reader, evidence:List[List[Tuple[int, str, int]]], claim:str):
        evidence_text: List[List[str]] = [[reader.get_doc_line(item[1],item[2]) for item in group] for group in evidence]
        flat_evidence_text: List[str] = self._flatten(evidence_text)

        # Key sets in dictionaries preserve insert order whereas python sets do not
        evidence_dict = dict()
        for item in flat_evidence_text:
            evidence_dict[item] = 1
        evidence = " ".join(evidence_dict.keys())

        return [{"evidence":evidence, "claim":claim, "evidence_group":None}]



import os
import random
class SimpleRandom():
    instance = None

    def __init__(self,seed):
        self.seed = seed
        self.random = random.Random(seed)

    def next_rand(self,a,b):
        return self.random.randint(a,b)

    @staticmethod
    def get_instance():
        if SimpleRandom.instance is None:
            SimpleRandom.instance = SimpleRandom(SimpleRandom.get_seed())
        return SimpleRandom.instance

    @staticmethod
    def get_seed():
        return int(os.getenv("RANDOM_SEED", 1234))


import json

from overrides import overrides
from typing import Dict, Union, Iterable, List, Tuple, Optional

from allennlp.data import Tokenizer, TokenIndexer, Instance
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import TextField, LabelField, MetadataField
from allennlp.data.token_indexers import SingleIdTokenIndexer
from allennlp.data.tokenizers import WordTokenizer

from fever.reader.document_database import FEVERDocumentDatabase
from fever.reader.preprocessing import FEVERInstanceGenerator, ConcatenateEvidence
from fever.reader.simple_random import SimpleRandom




@DatasetReader.register("fever")
class FEVERDatasetReader(DatasetReader):
    
    def __init__(self, database: Optional[Union[FEVERDocumentDatabase, str]] = "data/fever/fever.db",
                 wiki_tokenizer: Tokenizer = None,
                 claim_tokenizer: Tokenizer = None,
                 token_indexers: Dict[str, TokenIndexer] = None,
                 instance_generator: FEVERInstanceGenerator = None) -> None:

        super().__init__()
        if type(database) == str:
            database = FEVERDocumentDatabase(database)

        self._database = database
        self._wiki_tokenizer = wiki_tokenizer or WordTokenizer()
        self._claim_tokenizer = claim_tokenizer or WordTokenizer()
        self._token_indexers = token_indexers or {'tokens': SingleIdTokenIndexer()}
        self._instance_generator = instance_generator or ConcatenateEvidence()

    def get_doc_lines(self, page_title:str) -> List[str]:
        doc_lines = self._database.get_doc_lines(page_title)
        return [line.split('\t')[1] for line in doc_lines]

    def get_doc_line(self, page_title: str, line_number: int) -> str:
        if line_number is None:
            raise Exception("It looks like an NEI page is being loaded, but no evidence is present")

        if line_number > -1:
            return self.get_doc_lines(page_title)[line_number]
        else:
            return self.get_random_line(self.get_non_empty_lines(self.get_doc_lines(page_title)))

    def get_random_line(self,lines:List[str]) -> str:
        return lines[SimpleRandom.get_instance().next_rand(0, len(lines) - 1)]

    def get_non_empty_lines(self, lines:List[str]) -> List[str]:
        return [line for line in lines if len(line.strip())]

    @overrides
    def text_to_instance(self, claim_id:int, evidence_group:int, evidence:str, claim:str, label:str = None) -> Instance:

        claim_tokens = self._claim_tokenizer.tokenize(claim)
        evidence_tokens = self._wiki_tokenizer.tokenize(evidence)

        instance_meta = {"claim_id":claim_id,
                         "evidence_group": evidence_group,
                         "premise_tokens": evidence_tokens,
                         "hypothesis_tokens": claim_tokens,}

        instance_dict = {"premise": TextField(evidence_tokens, self._token_indexers),
                         "hypothesis": TextField(claim_tokens, self._token_indexers),
                         "metadata": MetadataField(instance_meta)
                        }

        if label is not None:
            instance_dict["label"] = LabelField(label)

        return Instance(instance_dict)

    def generate_instances(self,
                           claim_id:int,
                           evidence: List[List[Tuple[int,str,int]]],
                           claim: str,
                           label:str = None) -> Iterable[Instance]:

        generated = self._instance_generator.generate_instances(self, evidence, claim)
        return [self.text_to_instance(claim_id, item['evidence_group'], item['evidence'], item['claim'],label) for item in generated]

    @overrides
    def _read(self, file_path: str):
        with open(file_path,"r") as f:
            for line in f:
                instance = json.loads(line)

                claim_id:int = instance['id']
                claim:str = instance['claim']
                evidence: List[List[Tuple[int,int,str,int]]] = instance['evidence']
                evidence: List[List[Tuple[int, str,int]]] = [[(item[1], item[2], item[3]) for item in group] for group in evidence]

                label:str = instance['label'] if 'label' in instance else None

                yield from self.generate_instances(claim_id, evidence, claim, label)



from setuptools import setup, find_packages
import sys

with open('requirements.txt') as f:
    reqs = f.read()

reqs = reqs.strip().split('\n')

install = [req for req in reqs if not req.startswith("git+git://")]
depends = [req.replace("git+git://", "git+http://") for req in reqs if req.startswith("git+git://")]

setup(
    name='fever-allennlp-reader',
    version='1.6.39',
    author='James Thorne',
    author_email='james@jamesthorne.co.uk',
    url='https://jamesthorne.co.uk',
    description='Fact Extraction and VERification dataset reader for AllenNLP',
    long_description="readme",
    license=license,
    python_requires='>=3.6',
    package_dir={'fever': 'src/fever',
                 'fever.reader': 'src/fever/reader'},
    packages=['fever',
              'fever.reader'
              ],
    install_requires=install,
    dependency_links=depends,

)



import json

from allennlp.common import JsonDict
from allennlp.models import Model
from allennlp.predictors import Predictor
from overrides import overrides
from typing import Dict, Union, Iterable, List, Tuple, Optional

from allennlp.data import Tokenizer, TokenIndexer, Instance
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import TextField, LabelField, MetadataField
from allennlp.data.token_indexers import SingleIdTokenIndexer
from allennlp.data.tokenizers import WordTokenizer

from fever.reader.document_database import FEVERDocumentDatabase
from fever.reader.preprocessing import FEVERInstanceGenerator, ConcatenateEvidence
from fever.reader.simple_random import SimpleRandom


@Predictor.register("fever")
class FEVERPredictor(Predictor):
    
    def _json_to_instance(self, json_dict: JsonDict) -> Instance:
        claim_id: int = json_dict['id'] if "id" in json_dict else None
        claim: str = json_dict['claim']
        label: str = json_dict['label'] if 'label' in json_dict else None
        evidence: List[Tuple[str, int]] = json_dict['predicted_sentences']
        evidence: List[List[Tuple[str,int]]] = [[(None, item[0],item[1]) for item in evidence]]


        generated = self._instance_generator.generate_instances(self, evidence, claim)[0]

        return self._dataset_reader.text_to_instance(claim_id, None, generated["evidence"], claim, label)

    def predict(self, json_line:str) -> JsonDict:
        return self.predict_json(json.loads(json_line))

    def __init__(self, model: Model, dataset_reader: DatasetReader) -> None:
        super().__init__(model, dataset_reader)




import logging
import sqlite3
import unicodedata
from typing import Iterable

from allennlp.common.file_utils import cached_path


class FEVERDocumentDatabase(object):
    # An implementation inspired by the Facebook DrQA sqlite document database reader

    def __init__(self, path: str):
        self._connection = self.connect(cached_path(path))

    @staticmethod
    def connect(path:str):
        try:
            return sqlite3.connect(path)
        except sqlite3.Error as e:
            logging.exception(e)
            logging.critical("Unable to load sqlite database")
            raise e

    def get_doc_lines(self,page_title:str) -> str:
        cursor = self._connection.cursor()
        cursor.execute("SELECT lines FROM documents WHERE id = :page_title",
                       ({
                           "page_title":unicodedata.normalize("NFD",page_title,)
                       }))
        result = cursor.fetchone()
        cursor.close()

        if result is None:
            raise Exception("Document not found")

        return result[0].split("\n")
